# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Abstract interface for storage implemenations."""


class Storage:

    """Storage interface to implement.

    A Blob Storage represents a location where encrypted blobs can be stored
    and synced from/to.
    """

    def list_blob_ids(self):
        """Get the ids of all blobs in the storage."""
        raise NotImplementedError

    def contains(self, id_):
        """Return if the given blob id is present in the storage."""
        return id_ in self.list_blob_ids()

    def delete(self, id_):
        """Delete a given blob in the storage."""
        raise NotImplementedError

    def delete_everything(self, *, confirm=False):
        """Delete every blob from the storage."""
        for blob_id in self.list_blob_ids():
            self.delete(blob_id)

    def import_blob(self, id_, filename):
        """Add an encrypted blob file to the storage."""
        raise NotImplementedError

    def read_in_chunks(self, id_, chunk_size=1024):
        """Read a blob chunk by chunk."""
        raise NotImplementedError

    def missing_from(self, other_storage):
        """Get the ids present locally but not on an other storage."""
        blobs_self = set(self.list_blob_ids())
        blobs_other = set(other_storage.list_blob_ids())
        return blobs_self - blobs_other

    def push_to(self, other_storage):
        """Push local blobs to another storage."""
        raise NotImplementedError

    def push_blob_to(self, id_, other_storage):
        """Push a local blob to another storage."""
        raise NotImplementedError



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import os
import uuid
import hashlib
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Util import Counter

from .settings import TMPSTORAGE_PATH
from .utils import read_in_chunks

hashing = hashlib.sha256


def random_key():
    "Return a randomly generated AES key"
    random = Random.new()
    key = random.read(AES.key_size[2])  # 256 bits
    return key


def copy_and_encrypt(storage, filepath, key):
    "Encrypt the file into the data store, and returns its id."

    tmp_name = 'billabong-{}.part'.format(uuid.uuid4())
    tmp_destination = os.path.join(TMPSTORAGE_PATH, tmp_name)

    source_hash = hashing()
    enc_hash = hashing()
    ctr = Counter.new(128)
    crypto = AES.new(key, AES.MODE_CTR, counter=ctr)

    with open(filepath, 'rb') as source_file, \
            open(tmp_destination, 'wb') as dest_file:

        for chunk in read_in_chunks(source_file):
            source_hash.update(chunk)
            enc_chunk = crypto.encrypt(chunk)
            enc_hash.update(enc_chunk)
            dest_file.write(enc_chunk)

    # Now that we have the hash of the encrypted file, move the
    # encrypted file to the storage and return the hash.
    id_ = enc_hash.hexdigest()

    with open(tmp_destination, 'rb') as tmp_file:
        storage.import_blob(id_, tmp_file)
    os.remove(tmp_destination)

    return enc_hash.hexdigest()


def counter_for_offset(offset):
    "Return a Counter for the given offset."
    initial_value = 1 + (offset // 16)
    return Counter.new(128, initial_value=initial_value)


def decrypt_blob(storage, blob_id, key, offset=0, length=None):
    "Decrypt the content of a blob through a generator."

    modulo = offset % 16
    # Compute a file offset, might be lower than the real offset
    file_offset = offset - modulo if offset else 0

    end = offset + length if length else None
    ctr = counter_for_offset(offset)

    crypto = AES.new(key, AES.MODE_CTR, counter=ctr)

    chunks_generator = storage.read_in_chunks(blob_id, offset=file_offset,
                                              chunk_size=10)

    for i, enc_chunk in chunks_generator:
        chunk = crypto.decrypt(enc_chunk)

        if end is not None:
            here = offset - modulo + i*10
            if end <= here + 10:
                chunk = chunk[:end - here]
                yield chunk
                break

        if i == 0:
            yield chunk[modulo:]
        else:
            yield chunk



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import os
import json

from .storage import load_store
from .inventory import load_inventory

settings_path_candidates = [
    "./billabong.json",
    os.path.expanduser("~/.config/billabong.json"),
    "/etc/billabong.json",
]

for candidate in settings_path_candidates:
    if os.path.isfile(candidate):
        settings_path = candidate
        break
else:
    raise FileNotFoundError("Billabong configuration file not found.")

settings = json.load(open(settings_path))

TMPSTORAGE_PATH = settings.get('tmp_directory', '/tmp')

inventory = load_inventory(settings['inventory'])
stores = [load_store(r) for r in settings['stores']]



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
Check the integrity of the data.
"""

import logging
from base64 import b64decode

from .encryption import hashing, decrypt_blob
from .utils import read_in_chunks
from .exceptions import CheckError
from .settings import inventory, stores


def compute_hash(file_object, chunk_size=1024):
    """Compute the hash of the content of a file object using
    the given hashing function.
    """
    file_hash = hashing()
    for chunk in read_in_chunks(file_object, chunk_size):
        file_hash.update(chunk)
    return file_hash


def check_data(id=None, meta=None, raises=False):

    if id and not meta:
        meta = inventory.get_record(id)
    elif meta and not id:
        id = meta['id']
    else:
        raise ValueError("Missing value for 'id' or 'meta'.")

    blob_id = meta['blob']
    check_enc_data(blob_id)

    key = b64decode(meta['key'])
    hash = meta['hash']
    check_clear_data(blob_id, key, hash)


def check_enc_data(blob_id, raises=False):
    "Check the validity of an encrypted blob"

    enc_path = stores[0]._blob_path(blob_id)
    with open(enc_path, 'rb') as enc_file:
        enc_hash = compute_hash(enc_file)

    if id != enc_hash.hexdigest():
        if raises:
            raise CheckError(
                "Data does not match the hash for id '{}'".format(id))
        else:
            logging.error(
                "Data does not match the hash for id '{}'".format(id))


def check_clear_data(id, key, hash):
    "Check the validity of the clear data inside a blob"

    clear_data = decrypt_blob(stores[0], id, key)

    clear_hash = hashing()
    for chunk in clear_data:
        clear_hash.update(chunk)

    if hash != "sha256-" + clear_hash.hexdigest():
        raise CheckError()



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""Blob Storage."""

from .abstract import Storage
from .folder import FolderStorage
from .http import HTTPStorage
from .ssh import SSHStorage

assert Storage
assert FolderStorage
assert HTTPStorage
assert SSHStorage


def load_store(store_settings):
    """Instanciate a storage instance from settings."""
    type_ = store_settings['type']
    args = store_settings.get('args', {})

    if type_ == 'FolderStorage':
        return FolderStorage(**args)
    elif type_ == 'HTTPStorage':
        return HTTPStorage(**args)
    elif type_ == 'SSHStorage':
        return SSHStorage(**args)
    else:
        raise ValueError("Unknown type", type_)



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import datetime
import json


def json_handler(obj):
    if isinstance(obj, datetime.datetime) or isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        return obj.decode()
    else:
        print([obj])
        return json.JSONEncoder().default(obj)


def dumps(obj, indent=2):
    return json.dumps(obj, default=json_handler, indent=indent)


def json_loader(dico):
    for key, value in dico.items():
        if key == 'timestamp' and isinstance(value, str):
            try:
                dico[key] = datetime.datetime.strptime(value,
                                                       "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError as error:
                print("WARNING: ", error)
    return dico


def loads(string):
    return json.loads(string, object_hook=json_loader)


def read_in_chunks(file_object, chunk_size=1024):
    """Lazy function (generator) to read a file piece by piece.
    Default chunk size: 1k."""
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
Storage and manipulation of records containing metadata.
"""
import os
import os.path

from .utils import loads, dumps


class Inventory:
    """An Inventory represents a location where records of metadata can be stored
    and synced from/to.
    """

    def list_record_ids(self):
        raise NotImplementedError

    def delete(self, id_):
        raise NotImplementedError

    def get_record(self, id_):
        raise NotImplementedError

    def save_record(self, value):
        raise NotImplementedError

    def list_records(self):
        "Yield all records content."
        for id_ in self.list_record_ids():
            yield self.get_record(id_)

    def delete_everything(self, *, confirm=False):
        "Delete every record from the inventory"
        for record_id in self.list_record_ids():
            self.delete(record_id)

    def list_record_keyvalues(self, key):
        "List the value of one key for all records."
        for id_ in self.list_record_ids():
            record = self.get_record(id_)
            yield record['info'][key]

    def list_record_paths(self):
        "List all the paths of all records, used for 'ls' for example."
        yield from self.list_record_keyvalues('path')

    def list_record_filenames(self):
        "List all records by their filename, used for 'ls' for example."
        for value in self.list_record_keyvalues('filename'):
            yield value.replace('/', '.')

    def search(self, term):
        "Search for a term in the metadata and return the corresponding ids"
        for record in self.list_records():
            if term in dumps(record):
                yield record['id']

    def search_id(self, partial_id):
        "Search for an id given the first part of an id."
        for record_id in self.list_record_ids():
            if record_id.startswith(partial_id):
                yield record_id

    def id_from_filename(self, filename):
        "Return the id of the first record matching filename"
        for record in self.list_records():
            if filename == record['info']['filename']:
                return record['id']


class FolderInventory(Inventory):
    """Inventory in a folder accessible via the local filesystem.
    """

    def __init__(self, path):
        self.path = path

    def _record_path(self, id_):
        "Returns the path on the filesystem where a record is stored."
        return os.path.join(self.path, id_ + '.json')

    def list_record_ids(self):
        "List all record ids present in the inventory."
        for filename in os.listdir(self.path):
            id_ = filename[:-len('.json')]
            yield id_

    def delete(self, id_):
        "Delete a record from the inventory"
        os.remove(self._record_path(id_))

    def get_record(self, id_):
        "Load metadata for the given id."
        filepath = self._record_path(id_)
        return loads(open(filepath, 'r').read())

    def save_record(self, record):
        destination = self._record_path(record['id'])
        open(destination, 'w').write(dumps(record))


def load_inventory(settings):
    type_ = settings['type']
    args = settings.get('args', {})

    if type_ == 'FolderInventory':
        return FolderInventory(**args)



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os.path
from uuid import uuid4
from base64 import b64encode, b64decode
from datetime import datetime

import magic

from .encryption import random_key, copy_and_encrypt, decrypt_blob
from .check import compute_hash


class Billabong:

    """High-level interface above inventory and stores."""

    def __init__(self, inventory, stores):
        self.inventory = inventory
        self.stores = stores

    def add_file(self, filepath, *, key=None):
        """Import a file into Billabong and return the corresponding record."""
        # Resolve symlinks
        realpath = os.path.realpath(filepath)

        if not os.path.isfile(realpath):
            raise FileNotFoundError

        if key is None:
            key = random_key()

        with open(realpath, 'rb') as source_file:
            file_hash = compute_hash(source_file)

        storage = self.stores[0]
        blob_hash = copy_and_encrypt(storage, realpath, key)

        record = {
            'key': b64encode(key).decode(),
            'hash': 'sha256-' + file_hash.hexdigest(),
            'blob': blob_hash,
            'size': os.path.getsize(realpath),
            'timestamp': datetime.now(),
            'id': uuid4().hex,

            'info': {
                'type': magic.from_file(realpath).decode(),
                'mimetype': magic.from_file(realpath, mime=True).decode(),
                'filename': os.path.basename(filepath),
                'path': filepath,
                }
            }

        self.inventory.save_record(record)
        return record

    def get(self, id_):
        """Return a Record object from an id_."""
        return self.inventory.get_record(id_)

    def delete(self, id_):
        """Delete a Record and the corresponding blob."""
        blob_id = self.inventory.get_record(id_)['blob']
        for store in self.stores:
            try:
                store.delete(blob_id)
            except NotImplementedError:
                pass
        self.inventory.delete(id_)

    def read(self, id_, length=None, offset=0, chunk_size=1024):
        """Return data from the blob of this file."""
        record = self.inventory.get_record(id_)
        key = b64decode(record['key'])
        blob_id = record['blob']

        for store in self.stores:
            if store.contains(blob_id):
                return decrypt_blob(store, blob_id, key,
                                    offset=offset, length=length)
        else:
            raise FileNotFoundError



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Storage implementation based on the HTTP protocol."""


import os.path
import http.client
from urllib.parse import urlparse

from .abstract import Storage


class HTTPStorage(Storage):

    """Blob storage on a remote read-only HTTP(S) host."""

    def __init__(self, url):
        """Initialize an HTTP storage based on the given root url."""
        self.url = url

    def contains(self, id_):
        """Return if the given blob id is present in the storage."""
        u = urlparse(self.url)
        conn = http.client.HTTPConnection(host=u.hostname,
                                          port=u.port)
        conn.request('HEAD', os.path.join(u.path, id_))
        resp = conn.getresponse()
        return (resp.status == 200)


    def read_in_chunks(self, id_, offset=0, chunk_size=1024):
        """Read a blob file from HTTP chunk by chunk."""
        print('read http chunk {} {}'.format(id_, offset))

        u = urlparse(self.url)
        conn = http.client.HTTPConnection(host=u.hostname,
                                          port=u.port)
        conn.request('GET', os.path.join(u.path, id_),
                     headers={'Range': 'bytes={}-'.format(offset)})
        resp = conn.getresponse()

        if resp.status == 206:
            # 206 = Partial Content
            i = 0
            chunk = resp.read(chunk_size)
            while chunk:
                yield i, chunk
                chunk = resp.read(chunk_size)
                i += 1

        elif resp.status == 404:
            raise ValueError("404 not found")
        else:
            raise NotImplementedError("Range unsupported, status {}"
                                      .format(resp.status))



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
FUSE file system over the blob system.
"""

import errno

from fuse import FUSE, FuseOSError, Operations

from billabong import billabong
from billabong.settings import inventory, settings


def id_from_path(path):
    "Get the blob id from a FUSE path"
    if path.startswith('/blobs/'):
        id_ = path[len('/blobs/'):]
    elif path.startswith('/files/'):
        filename = path.split('/')[-1]
        id_ = inventory.id_from_filename(filename)
    else:
        id_ = None

    if id_:
        return id_
    else:
        raise FuseOSError(errno.ENOENT)


class BillabongFilesystem(Operations):

    def __init__(self):
        pass

    # ----- File methods -----

    def getattr(self, path, fh=None):
        if path in ('/', '/blobs', '/files'):
            return dict(
                st_mode=16877,
                st_ino=7077891,
                st_dev=64514,
                st_nlink=37,
                st_uid=1000,
                st_gid=1000,
                st_size=4096,
                st_atime=1428702934,
                st_mtime=1428702934,
                st_ctime=1428702934,
            )
        else:
            id_ = id_from_path(path)
            print('id_', path, id_)
            meta = inventory.get_record(id_)
            return dict(
                st_mode=33188,
                st_ino=7093655,
                st_dev=64514,
                st_nlink=1,
                st_uid=1000,
                st_gid=1000,
                st_size=meta['size'],
                st_atime=1428701319,
                st_mtime=1428514886,
                st_ctime=1428514886
            )

    def readdir(self, path, fh):
        if path == '/':
            return ['blobs', 'files']
        elif path == '/blobs':
            return inventory.list_record_ids()
        elif path == '/files':
            return inventory.list_record_filenames()
        else:
            print('Unknown path', path)

    def read(self, path, length, offset, fh):
        if path.startswith('/blobs/'):
            id_ = id_from_path(path)
            print(path, length, offset, fh)
            data = billabong.read(id_, offset=offset, length=length)
            return b''.join(data)
        elif path.startswith('/files/'):
            id_ = id_from_path(path)
            if id_:
                return b''.join(
                    billabong.read(id_, offset=offset, length=length))
            else:
                raise FuseOSError(errno.ENOENT)


def mount_fuse(path=None, foreground=True):
    if path is None:
        path = settings['mount']

    FUSE(BillabongFilesystem(), path, foreground=foreground)



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
# Billabong

A reliable, distributed, encrypted storage and backup solution for all of
your photos, videos, music and other static data.
"""

import os
import sys
from baker import command, run

try:
    from pygments import highlight
    from pygments.lexers import JsonLexer
    from pygments.formatters import TerminalFormatter
except ImportError:
    highlight = None

from billabong import billabong
from billabong.settings import inventory
from billabong.check import check_data, check_enc_data
from billabong.utils import dumps
from billabong.sync import push_blobs, pull_blobs


def print_record(record):
    if highlight and '--no-color' not in sys.argv:
        print(highlight(dumps(record),
                        JsonLexer(),
                        TerminalFormatter()))
    else:
        print(dumps(record))


@command
def ls():
    format_ = "{:>8} {:>8} {:>8}"
    for r in inventory.list_records():
        print(format_.format(r['id'][:8],
                             r['size'],
                             r['info']['filename']))


@command
def blobs():
    for i in inventory.list_record_ids():
        print(i)


@command
def add(*targets):
    for target in targets:
        record = billabong.add_file(target)
        print_record(record)


@command
def get(id_):
    raise NotImplemented


@command
def info(*ids):
    for id_ in ids:
        meta = inventory.get_record(id_)
        print_record(meta)


@command
def echo(id_):
    data = billabong.read(id_)

    # Write bytes to stdout:
    fp = os.fdopen(sys.stdout.fileno(), 'wb')
    for chunk in data:
        fp.write(chunk)
        fp.flush()


@command
def search(term):
    for i in inventory.search(term):
        print(i)


@command
def check():
    "Check the validity of all blobs and metadata"
    for i in inventory.list_record_ids():
        check_data(i)
        check_enc_data(i)


@command
def push():
    "Push blobs to sync storage"
    push_blobs()


@command
def pull():
    "Pull blobs from sync storage"
    pull_blobs()


@command
def status():
    "Print a global status of the inventory and storage."
    print("Inventory:")
    print("  {:>4} records".format(
          len(list(inventory.list_record_ids()))))
    print("  {:>4} bytes total".format(
          sum(i['size'] for i in inventory.list_records())))


@command
def mount(path=None, foreground=False):
    from billabong.fs import mount_fuse
    mount_fuse(path, foreground)


@command
def version():
    from billabong import __version__
    print(__version__)


if __name__ == '__main__':
    run()



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
# Billabong

A reliable, distributed, encrypted storage and backup solution for all of
your photos, videos, music and other static data.
"""

from .settings import inventory, stores
from .core import Billabong

__author__ = "Hugo Herter http://hugoherter.com"
__version__ = "0.2.2"


billabong = Billabong(inventory, stores)



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Storage implementation based on the filesystem."""

import os
from shutil import copyfileobj

from billabong.utils import read_in_chunks
from .abstract import Storage


class FolderStorage(Storage):

    """Blob Storage in a folder accessible via the local filesystem."""

    def __init__(self, path):
        """Initialize a Folder storage based on the given path."""
        self.path = path

    def _blob_path(self, id_):
        """Get the path on the filesystem where a blob is stored."""
        return os.path.join(self.path, id_)

    def list_blob_ids(self):
        """List all ids present in the blobs storage."""
        for id_ in os.listdir(self.path):
            yield id_

    def delete(self, blob_id):
        """Delete a blob from the storage."""
        os.remove(self._blob_path(blob_id))

    def import_blob(self, id_, blobfile):
        """Add an encrypted blob file to the storage by copying the file."""
        copyfileobj(blobfile,
                    open(self._blob_path(id_), 'wb'))

    def read_in_chunks(self, id_, offset=0, chunk_size=1024):
        """Read a blob file chunk by chunk."""
        path = self._blob_path(id_)
        fd = open(path, 'rb')
        fd.seek(offset)
        return enumerate(read_in_chunks(fd, chunk_size))

    def push_to(self, other_storage):
        """Push local blobs to another storage."""
        for id_ in self.missing_from(other_storage):
            self.push_blob_to(id_, other_storage)

    def push_blob_to(self, id_, other_storage):
        """Push a local blob to another storage."""
        blobfile = open(self._blob_path(id_), 'rb')
        other_storage.import_blob(id_, blobfile)



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from shutil import copy2
from .settings import stores


def push_blobs():
    for store in stores[1:]:
        try:
            stores[0].push_to(store)
        except NotImplementedError:
            print("Push not implemented for '{}'".format(store))


def push_blob(blob_id, storage, remote):
    print('pushing blob', blob_id)
    path_local = stores[0]._blob_path(blob_id)
    path_remote = remote._blob_path(blob_id)
    copy2(path_local, path_remote)


def pull_blobs():
    for store in stores[1:]:
        try:
            store.push_to(stores[0])
        except NotImplementedError:
            print("Pull not implemented for '{}'".format(store))



# Copyright (c) 2015 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Storage implementation based on the SSH protocol."""

import os.path
import paramiko
from .abstract import Storage

from billabong.utils import read_in_chunks


class SSHStorage(Storage):

    """Blob storage on a remote read-write SSH host."""

    def __init__(self, path, host, port=22):
        """Initialize an SSH storage based on the given credentials."""
        self.path = path
        self.host = host
        self.port = port

    def _blob_path(self, id_):
        """Get the path on the filesystem where a blob is stored."""
        return os.path.join(self.path, id_)

    def _ssh_connection(self):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.host, port=self.port)
        return ssh

    def list_blob_ids(self, read_aheads=50):
        """List all ids present in the blobs storage."""
        try:
            ssh = self._ssh_connection()
            sftp = ssh.open_sftp()
            for i in sftp.listdir_iter(self.path, read_aheads=read_aheads):
                yield i.filename
            sftp.close()
        finally:
            ssh.close()

    def import_blob(self, id_, blobfile, callback=None):
        """Add an encrypted blob file to the storage by copying the file."""
        path = self._blob_path(id_)

        try:
            ssh = self._ssh_connection()
            sftp = ssh.open_sftp()
            sftp.putfo(blobfile, path, callback=callback, confirm=True)
            sftp.close()
        finally:
            ssh.close()

    def read_in_chunks(self, id_, offset=0, chunk_size=1024):
        """Read a blob file chunk by chunk."""
        path = self._blob_path(id_)

        try:
            ssh = self._ssh_connection()
            sftp = ssh.open_sftp()

            fd = sftp.file(path, mode='r')
            fd.seek(offset)

            i = 0
            for chunk in read_in_chunks(fd, chunk_size):
                yield (i, chunk)

            sftp.close()
        finally:
            ssh.close()



#!/usr/bin/env python3
# coding: utf-8

# Copyright (c) 2014 "Hugo Herter http://hugoherter.com"
#
# This file is part of Billabong.
#
# Intercom is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

setup(name='Billabong',
      version='0.2.2',
      description='Personal Encrypted file storage, backup and sharing',
      long_description=open('README.rst').read(),
      author='Hugo Herter',
      author_email='@hugoherter.com',
      url='https://github.com/hoh/Billabong/',
      packages=['billabong', 'billabong.storage'],
      install_requires=['pycrypto', 'python-magic', 'baker', 'paramiko'],
      license='AGPLv3',
      keywords="encrypted storage backup distributed",
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Programming Language :: Python :: 3',
                   'Intended Audience :: Developers',
                   'Intended Audience :: End Users/Desktop',
                   'Intended Audience :: Information Technology',
                   'Natural Language :: English',
                   'Operating System :: POSIX :: Linux',
                   'License :: OSI Approved :: GNU Affero General Public License v3',
                   'Topic :: Database',
                   'Topic :: Communications :: File Sharing',
                   'Topic :: Desktop Environment :: File Managers',
                   'Topic :: System :: Archiving :: Backup',
                   'Topic :: System :: Filesystems',
                   ],
      )
